/*
 * acat.cpp:
 *
 * cat the contents of an AFF file...
 */

/*
 * Copyright (c) 2005
 *	Simson L. Garfinkel and Basis Technology, Inc. 
 *      All rights reserved.
 *
 * This code is derrived from software contributed by
 * Simson L. Garfinkel
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by Simson L. Garfinkel
 *    and Basis Technology Corp.
 * 4. Neither the name of Simson Garfinkel, Basis Technology, or other
 *    contributors to this program may be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SIMSON GARFINKEL, BASIS TECHNOLOGY,
 * AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL SIMSON GARFINKEL, BAIS TECHNOLOGy,
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.  
 */


#include "afflib.h"
#include "afflib_i.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <zlib.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <assert.h>

char *progname = "acat";
int  *segments = 0;
int  num_segments = 0;
int  opt_info     = 0;
char *opt_segname=0;
int  opt_pagenum = -1;


void usage()
{
    printf("usage: %s [options] infile [outfile]\n",progname);
    printf("options:\n");
    //printf("     -s segname   --- just output segment 'segname'\n");
    printf("     -p ###  --- just output page number ###\n");
    exit(0);
}

void add_segment(int i)
{
    segments = (int *)realloc(segments,sizeof(int)*(num_segments+1));
    segments[num_segments] = i;
    num_segments++;
}


int compar(const void *a_,const void *b_)
{
    int a = *(int *)a_;
    int b = *(int *)b_;
    if(a<b) return -1;
    if(a>b) return 1;
    return 0;
}

int print_page(AFFILE *af,FILE *outfile,int pagenum)
{
    unsigned char *buf = (unsigned char *)malloc(af->image_pagesize);
    if(buf==0){
	err(1,"malloc(%d) failed",(int)af->image_pagesize);
    }
    uint64 offset = pagenum * af->image_pagesize; // go to that location
    af_seek(af,offset,SEEK_SET);
    int bytes = af_read(af,buf,af->image_pagesize); // read what we can
    fwrite(buf,1,bytes,outfile);	// send to the output
    free(buf);
    return 0;
}


int main(int argc,char **argv)
{
    int bflag, ch;
     const char *infile;

     progname = argv[0];
     segments = (int *)malloc(0);
     bflag = 0;
     while ((ch = getopt(argc, argv, "s:h?vip:")) != -1) {
             switch (ch) {
	     case 's':
		 opt_segname = optarg;
		 break;
	     case 'p':
		 opt_pagenum = atoi(optarg);
		 break;
	     case 'h':
             case '?':
             default:
		 usage();
	     }
     }
     argc -= optind;
     argv += optind;

     if(argc<1){
	 usage();
     }

     infile = *argv++;
     argc--;

     const char *outfilename = 0;
     if(argc>0){
	 outfilename = *argv++;
	 argc--;
     }

     FILE *outfile = stdout;

     if(outfilename){
	 outfile = fopen(outfilename,"wb");
	 if(!outfile) err(1,outfilename);
     }


     /* Read all of the segments from beginning to end and capture
      * all the segment numbers...
      */
     AFFILE *af = af_open(infile,O_RDONLY,0);
     if(!af) err(1,infile);


     if(opt_pagenum != -1){
	 print_page(af,outfile,opt_pagenum);
	 goto done;
     }

     af_rewind_seg(af);			// start at the beginning
     char segname[AF_MAX_NAME_LEN];
     memset(segname,0,sizeof(segname));
     while(af_get_next_seg(af,segname,sizeof(segname),0,0,0)==0){
	int page_num = af_segname_page_number(segname);
	if(page_num>=0) add_segment(page_num);
     }
     qsort(segments,num_segments,sizeof(int),compar);
	
     af_demand_sizes(af);		// better have it

     /* Now I have a list of segments; cat each one */
     for(int i=0;i<num_segments;i++){
	 print_page(af,outfile,segments[i]);
     }

 done:;

     af_close(af);
     fclose(outfile);
}
